
import java.awt.*;

/**
 * PSystem - DESCRIPTION
 * <p/>
 * An extension to the Processing project - http://Proce55ing.net
 * <p/>
 * Copyright (c) 2004 Simon Greenwold
 * Created: Feb 11, 2004
 * <p/>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * <p/>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * <p/>
 * You should have received a copy of the GNU Lesser General
 * Public License along with the Processing project; if not,
 * write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

public class PSystem implements MPConstants, Plugin {


  private static final float DEFAULT_GX = 0.0f;
  private static final float DEFAULT_GY = 0.03f;
  private static final float DEFAULT_GZ = 0.0f;

  private static final float DEFAULT_DRAG = 0.03f;

  private static final float DEFAULT_SPRING_DAMPING = 0.01f;
  private static final float DEFAULT_SPRING_STRENGTH = 0.005f;
  private static final float DEFAULT_SPRING_REST_LENGTH = 18.0f;

  private static final float DEFAULT_TIME_STEP = 0.5f;

  private static final float DEFAULT_MASS = 1.0f;
  private static final float DEFAULT_COLLISION_RADIUS = 10.0f;
  private static final float DEFAULT_MIN_COLLISION_FORCE = 1.0f;
  private static final float DEFAULT_MAX_COLLISION_FORCE = 2.0f;

  float time;  // accumulated system time

  int nForces;
  int nParticles;
  boolean drawForces = true;
  Particle particles[] = null;
  Force forces[] = null;

  Surface surfaces[] = null;

  float gX, gY, gZ;
  float drag, timeStep;
  float defaultMass;
  float defaultSpringDamping;
  float defaultSpringRestLength;
  float defaultSpringStrength;
  float defaultCollisionRadius;
  float defaultMinCollisionForce;
  float defaultMaxCollisionForce;

  void initGlobals() {
    gX = DEFAULT_GX;
    gY = DEFAULT_GY;
    gZ = DEFAULT_GZ;
    drag = DEFAULT_DRAG;
    timeStep = DEFAULT_TIME_STEP;

    defaultMass = DEFAULT_MASS;
    defaultSpringDamping = DEFAULT_SPRING_DAMPING;
    defaultSpringStrength =  DEFAULT_SPRING_STRENGTH;
    defaultSpringRestLength = DEFAULT_SPRING_REST_LENGTH;
    defaultCollisionRadius = DEFAULT_COLLISION_RADIUS;
    defaultMinCollisionForce = DEFAULT_MIN_COLLISION_FORCE;
    defaultMaxCollisionForce = DEFAULT_MAX_COLLISION_FORCE;
  }

  private void preloop() {
    update();
  }

  private void postloop() {

  }

  public void draw() {
    drawParticles();
    drawForces();
  }

  public void drawParticles() {
    for (int i = 0; i < nParticles; i++) {
      if (particles[i].visible)
        particles[i].draw();
    }
  }

  public void drawForces() {
    if (!drawForces) return;
    for (int i = 0; i < nForces; i++) {
      if (forces[i].visible)
        forces[i].draw();
    }
  }

  public void update() {
    if (nParticles > 0) {
      zeroForces();
      applyForces();
      calcNextState();
      updateState();
    }

    if (surfaces != null) {
      for (int i = 0; i < surfaces.length; i++) {
        surfaces[i].update();
      }
    }
  }

  void drawSprings() {
    for (int i = 0; i < nForces; i++) {
      if (forces[i] instanceof Spring) {
        ((Spring)forces[i]).draw();
      }
    }
  }

  void addParticle(Particle p) {
    p.pSystem = this;
    if (particles == null) {
      particles = new Particle[START_PARTICLES];
    }

    if (nParticles >= particles.length) {
      Particle[] newParticles = new Particle[nParticles + START_PARTICLES];
      System.arraycopy(particles, 0, newParticles, 0, nParticles);
      particles = newParticles;
    }

    particles[nParticles] = p;
    p.init();
    if (p.collisionForce != null)
      addForce(p.collisionForce);
    nParticles++;
  }

  void removeParticle(int i) {
    System.arraycopy(particles, i + 1, particles, i, (nParticles - i) - 1);
    particles[nParticles] = null;
    nParticles--;
  }

  void removeParticle(Particle p) {
    for (int i = 0; i < nParticles; i++) {
      if (particles[i] == p) {
        for (int f = 0; f < nForces; f++) {
          if (forces[f] instanceof Spring) {
            if (((Spring)forces[f]).a == p || ((Spring)forces[f]).b == p) {
              removeForce(f);
            }
          }
          else if (forces[f] instanceof Magnet) {
            if (((Magnet)forces[f]).a == p) {
              removeForce(f);
            }
          }
          else if (forces[f] instanceof CollisionForce) {
            if (((CollisionForce)forces[f]).a == p) {
              removeForce(f);
            }
          }
        }
        removeParticle(i);
      }
    }
  }
  
  Magnet addMagnet(Particle a) {
    Magnet m = new Magnet(a);
    addForce(m);
    return m;
  }
  
  Magnet addMagnet(Particle a, float strength) {
    Magnet m = new Magnet(a);
    m.strength = strength;
    addForce(m);
    return m;
  }
  
  Spring addSpring(Particle a, Particle b) {
    Spring s = new Spring(a, b);
    addForce(s);
    return s;
  }
  
  Spring addSpring(Particle a, Particle b, float strength, float damping, float length) {
    Spring s = new Spring(a, b);
    addForce(s);
    s.strength = strength;
    s.restLength = length;
    s.damping = damping;
    return s;
  }
  
  void addForce(Force f) {
    if (forces == null) {
      forces = new Force[START_FORCES];
    }

    if (nForces >= forces.length) {
      Force[] newForces = new Force[nForces + START_FORCES];
      System.arraycopy(forces, 0, newForces, 0, nForces);
      forces = newForces;
    }

    forces[nForces] = f;
    nForces++;
    f.init(this);
  }

  void removeForce(int i) {
    System.arraycopy(forces, i + 1, forces, i, (nForces - i) - 1);
    forces[nForces] = null;
    nForces--;
  }

  void removeForce(Force f) {
    for (int i = 0; i < nForces; i++) {
      if (forces[i] == f) {
        removeForce(i);
      }
    }
  }

  void zeroForces() {
    for (int i = 0; i < nParticles; i++) {
      particles[i].clearForce();
    }
  }

  void applyForces() {
    Force force = null;
    for (int i = 0; i < nForces; i++)  {
      force = forces[i];
      if (force.active)
        force.applyForce();
    }
    Particle p = null;
    for (int i = 0; i < nParticles; i++)  {
      p = particles[i];
      if (! p.fixed()) {
        p.addForce(-drag * p.velocity[0], -drag * p.velocity[1], -drag * p.velocity[2]);
        //System.out.println("Adding GY " + globals[GY]);
        p.addForce(gX * p.mass, gY * p.mass, gZ * p.mass);
      }
    }
  }

  void applyForces(Particle p) {
    if (p.fixed()) return;

    Force force = null;
    for (int i = 0; i < nForces; i++)  {
      force = forces[i];
      if (force.active)
        force.applyForce(p);
    }

    p.addForce(-drag * p.velocity[0], -drag * p.velocity[1], -drag * p.velocity[2]);
    p.addForce(gX, gY, gZ);
  }

  void calcNextState() {
    calcNextState(timeStep);
  }

  void calcNextState(float step) {
    Particle p = null;
    for (int i = 0; i < nParticles; i++)  {
      p = particles[i];
      if (! p.fixed())
        p.calculateNext(step);
    }
  }

  void updateState() {
    Particle p = null;
    for (int i = 0; i < nParticles; i++)  {
      p = particles[i];
      if (! p.fixed())
        p.updateNext();
    }
  }

  public BApplet getBApplet() {
    return bApplet;
  }

  private BApplet bApplet;

  public String getVersion() {
    return "0.1a";
  }

  public int getMinCompatibleProcessingVersion() {
    return 68;
  }

  public String getDescription() {
    return "Particle and spring-mass systems for Processing.";
  }

  public void initialize(BApplet bApplet) {
    this.bApplet = bApplet;

    // This is to prevent a nasty hang that occurs on Windows sometimes.
    // I don't know why it works or if it's the only solution
    // or if it always works.
    if (!bApplet.online()) {
      Frame frame = new Frame();
      frame.setVisible(true);
      frame.setVisible(false);
    }

    bApplet.registerCall(PRE_LOOP_HOOK, new Runnable() {
      public void run() {
        preloop();
      }
    });
    bApplet.registerCall(POST_LOOP_HOOK, new Runnable() {
      public void run() {
        postloop();
      }
    });

    nParticles = 0;
    nForces = 0;
    initGlobals();
  }
  
  public void setGravity(float gY) {
    this.gY = gY;
  }
  
  public void setGravity(float gX, float gY, float gZ) {
    this.gX = gX;
    this.gY = gY;
    this.gZ = gZ;
  }


  private void addSurface(Surface s) {
    s.pSystem = this;
    if (surfaces == null) {
      surfaces = new Surface[1];
    }
    else {
      Surface[] newSurfaces = new Surface[surfaces.length + 1];
      System.arraycopy(surfaces, 0, newSurfaces, 0, surfaces.length);
      surfaces = newSurfaces;
    }

    surfaces[surfaces.length - 1] = s;
  }

  public Surface loadSurface(String surfaceName) {
    Surface s = new Surface(bApplet, this, surfaceName);
    addSurface(s);
    return s;
  }

}
